// $Id: TransparentWnd.pas,v 1.4 2007/04/05 07:51:00 q002sa Exp $

(*+++++ Transparent Windows +++++++ V 2.0 ++++++++++++++++++++++++++++++++++++++

For learnig, testing and commercial use

?***********************************************?
? by Michael Glitzner (michael.glitzner@aon.at) ?
?***********************************************?

* ABOUT THIS COMPONENT *
This component should capsule the new Windows 2000 API calls for Alpha Blending
(transparency). I think you can use this component in nearly any Delphi
(tested in BD5) version with Win2000 or higher (e.g. Windows Whistler).
Install this component in Delphi 5 with |COMPONENT|INSTALL COMPONENT|. Then you
have to choose the "transparentWnd.pas" file. After this the component is
installed in |WIN32|.
Only drop the component on the Form and use the prefered method to set
transparency.

V1.4
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+) SetLayeredWindowAttributes dynamically loaded at runtime (otherwise an error
   has occured in win9x)
+) SetWindowLongA and GetWindowLongA declarations removed (already declared in
   windows.pas)
~~~~~~~~~~~~~~~~~~~~~" Thanks to Budi Sukmawan for these improvements "~~~~~~~~~
+) SetOpaque
+) SetOpaqueHWND

V2.0
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+) Changed Longint's into THandle
+) Added some safeguards on percentage value and to handles (check wnd/no wnd)
+) Test Owner to be a TWinControl Descendant instead of blindly casting it to
   TForm, this allows you to create the controls dynamically.
+) Added Constructor and Destructor.
+) Variable comments
~~~~~~~~~~~~~~~~~~~~~" Thanks to wvd_vegt@knoware.nl (13-dec-2000) "~~~~~~~~~~~~
+) Corrected bug in SetTransparentHWND(): 0% is now inverted (not 100% transparent)
+) Redesigned SetTransparentHWND(). Because now library is loaded in constructor
   I rewrote this method.
+) Property AutoOpaque: When AutoOpaque has the value "true" then the component
   automatically sets the window opaque when the transparency percentage is 0%.
+) NOW IT SHOULD BE REALLY FAST

Derived from:
TComponent

Properties:
Percent
= Percentage of transparency ( only in use with SetTransparent() )

AutoOpaque
= When AutoOpaque has the value "true" then the component automatically sets the
  window opaque when the transparency percentage is 0%.

Inherited properties:
Properties from TComponent

Methods:
Create(AOwner)
= The constructor. Normally when you drag the component at your form in design
  mode you don't need it. But you can also create a component at runtime by
  using this constructor. It is also required to specify the owner of the comp.

Destroy()
= The destructor. You have to call it to destroy the component at runtime.

SetTransparent()
= Set the transparency of the parent window to the property percent.

SetTransparent(percent)
= Set the transparency of the parent window to percent(variable).

SetTransparentHWND(hwnd, percent)
= Set the transparency of the window with the hwnd to percent(variable).

SetOpaque()
= Switch of window transparency and turn window opaque (!!! is not the same as
  window transparency 0% !!!)

SetOpaqueHWND()
=  Switch of window transparency and turn window with hwnd opaque

Events:
Events from TComponent
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*)

unit TransparentWnd;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses
{$IFNDEF FPC}
  Windows,
{$ELSE}
  Windows, LCLIntf, LCLType, LMessages,
{$ENDIF}
  Messages, Classes, Controls, Forms;

type

  TPercentage = 0..100;

  TTransparentWnd = class(TComponent)
  private
    { Private declarations }
    fPercent : TPercentage;
    fAuto : boolean;
    fUser32: HMODULE;
    function GetClickThrough: boolean;
    procedure SetClickThrough(Value: boolean);
    function GetEnabled: boolean;
  protected

    { Protected declarations }
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    //These work on the Owner (a TWinControl decendant is the Minumum)
    procedure SetTransparent; overload;
    procedure SetTransparent(percent : TPercentage); overload;
    //These work on a Handle
    procedure SetTransparentHWND(hwnd: THandle; percent : TPercentage);
    procedure SetOpaqueHWND(hwnd : THandle);
    function GetClickThroughWnd(Handle: THandle): boolean;
    procedure SetClickThroughWnd(Handle: THandle; Value: boolean);

    procedure SetOpaque;

    property Enabled: boolean read GetEnabled;
    { Public declarations }
  published

    //This works on the Owner (a TWinControl decendant is the Minumum)
    property Percent: TPercentage read fPercent write fPercent default 0;
    property ClickThrough: boolean read GetClickThrough write SetClickThrough default false;
    property AutoOpaque: boolean read fAuto write fAuto default false;
    { Published declarations }
  end;

  procedure Register;

implementation

const LWA_ALPHA = $2;
const GWL_EXSTYLE = (-20);
const WS_EX_LAYERED = $80000;
const WS_EX_TRANSPARENT = $20;

var SetLayeredWindowAttributes: function (hwnd: LongInt; crKey: byte; bAlpha: byte; dwFlags: LongInt): LongInt; stdcall;

constructor TTransparentWnd.Create(AOwner: TComponent);
begin
  inherited;
  fUser32 := LoadLibrary('USER32.DLL');
  if fUser32 <> 0 then
    @SetLayeredWindowAttributes := GetProcAddress(fUser32, 'SetLayeredWindowAttributes')
  else
    SetLayeredWindowAttributes := nil;
end;

destructor TTransparentWnd.Destroy;
begin
  if fUser32 <> 0 then FreeLibrary(fUser32);
  inherited;
end;

procedure TTransparentWnd.SetOpaqueHWND(hwnd: THandle);
var
  old: THandle;
begin
  if IsWindow(hwnd) then begin
    old := GetWindowLongA(hwnd,GWL_EXSTYLE);
    SetWindowLongA(hwnd, GWL_EXSTYLE, old and not WS_EX_LAYERED);
  end;
end;

procedure TTransparentWnd.SetOpaque;
begin
  Self.SetOpaqueHWND((Self.Owner as TWinControl).Handle);
end;

procedure TTransparentWnd.SetTransparent;
begin
  Self.SetTransparentHWND((Self.Owner as TWinControl).Handle, fPercent);
end;

procedure TTransparentWnd.SetTransparentHWND(hwnd: THandle; percent : TPercentage);
var
  old: THandle;
begin
  if (fUser32 <> 0) and (Assigned(SetLayeredWindowAttributes)) and (IsWindow(hwnd)) then begin
    if (fAuto) and (percent = 0) then
      SetOpaqueHWND(hwnd)
    else begin
        fPercent := 100 - percent;
        old := GetWindowLongA(hwnd, GWL_EXSTYLE);
        SetWindowLongA(hwnd,GWL_EXSTYLE, old or WS_EX_LAYERED);
        SetLayeredWindowAttributes(hwnd, 0, (255 * fPercent) div 100, LWA_ALPHA);
      end;
    end;
end;

procedure TTransparentWnd.SetTransparent(percent: TPercentage);
begin
  Self.SetTransparentHWND((Self.Owner as TForm).Handle, percent);
end;

function TTransparentWnd.GetClickThrough: boolean;
begin
  Result := GetClickThroughWnd((Self.Owner as TWinControl).Handle);
end;

procedure TTransparentWnd.SetClickThrough(Value: boolean);
begin
  SetClickThroughWnd((Self.Owner as TWinControl).Handle, Value);
end;

function TTransparentWnd.GetClickThroughWnd(Handle: THandle): boolean;
var
  s: DWORD;
begin
  s := GetWindowLong(Handle, GWL_EXSTYLE);
  Result := s and WS_EX_TRANSPARENT = WS_EX_TRANSPARENT;
end;

procedure TTransparentWnd.SetClickThroughWnd(Handle: THandle; Value: boolean);
var
  s: DWORD;
begin
  s := GetWindowLong(Handle, GWL_EXSTYLE);
  if Value then
    SetWindowLong(Handle, GWL_EXSTYLE, s or WS_EX_TRANSPARENT)
  else
    SetWindowLong(Handle, GWL_EXSTYLE, s and not WS_EX_TRANSPARENT);
end;

procedure Register;
begin
  RegisterComponents('Win32', [TTransparentWnd]);
end;

function TTransparentWnd.GetEnabled: boolean;
begin
  Result := (fUser32 <> 0) and (Assigned(SetLayeredWindowAttributes));
end;

end.
