{
  MACAL (Mass Calculator)
  Copyright (C) 2007, Innsbruck Medical University

  Author: Stefan Ascher <stefan.ascher@i-med.ac.at>

  $Id: uSequence.pas,v 1.6 2007/04/05 07:51:00 q002sa Exp $
}

unit uSequence;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses
  SysUtils, classes, uMods, uMasses;

type
  EUnknownFileFormat = class(Exception)
  end;
  EInvalidSequence = class(Exception)
  end;

  TAminoAcid = class(TMass)
  private
    fThreeLetter: string;
  public
    property ThreeLetter: string read fThreeLetter write fThreeLetter;
  end;

  TAminoAcidList = class(TMassList)
  private

  protected
    function GetMass: double; virtual;
  public
    procedure LoadFromFile(const AFilename: string); override;
    property Mass: double read GetMass;
  end;

  TSequence = class(TAminoAcidList)
  private
    fMods: TModList;
    fAAs: TAminoAcidList;
    fName: string;
    function GetSequence: string;
    procedure SetSequence(Value: string);
    function GetLength: integer;
    procedure SetMassType(Value: TMassType);
    function GetName: string;
  protected
    function GetMass: double; override;
  public
    constructor Create;
    property Sequence: string read GetSequence write SetSequence;
    property Length: integer read GetLength;
    property AAs: TAminoAcidList read fAAs;
    property Mods: TModList read fMods;
    property MassType: TMassType read fMassType write SetMassType;
    property Name: string read GetName write fName;
    destructor Destroy; override;
    function GetAAPos(const p: integer): integer;
    procedure LoadFromFasta(const AFilename: string);
    procedure SaveToFasta(const AFilename: string);
    function GetAAs: string;
  end;

const
  AAS_FILE = 'aas.ini';

implementation

uses
  Math;

{ TAminoAcid }

{ TAminoAcidList }

procedure TAminoAcidList.LoadFromFile(const AFilename: string);
var
  sl: TStringList;
  i, p1: integer;
  str, str2: string;
  aa: TAminoAcid;
begin
  fFilename := AFilename;
  DeleteAll;
  if not FileExists(AFilename) then Exit;
  sl := TStringList.Create;
  try
    sl.LoadFromFile(AFilename);
    for i := 0 to sl.Count - 1 do begin
      str := Trim(sl[i]);
      if (str <> '') and (str[1] <> ';') then begin
        aa := TAminoAcid.Create(Self);
        Add(aa);
        aa.NameChar := str[1];

        System.Delete(str, 1, 2);
        p1 := Pos(';', str);
        if (p1 = 0) then
          Continue;
        str2 := Copy(str, 1, p1-1);
        aa.ThreeLetter := str2;

        System.Delete(str, 1, p1);
        p1 := Pos(';', str);
        if (p1 = 0) then
          Continue;
        str2 := Copy(str, 1, p1-1);
        aa.Name := str2;

        System.Delete(str, 1, p1);
        p1 := Pos(';', str);
        if (p1 = 0) then
          Continue;
        str2 := Copy(str, 1, p1-1);
        aa.Composition := str2;

        System.Delete(str, 1, p1);
        p1 := Pos(';', str);
        if (p1 = 0) then
          Continue;
        str2 := Copy(str, 1, p1-1);
        aa.MassMono := StrToMass(str2);

        System.Delete(str, 1, p1);
        if (str = '') then
          Continue;
        aa.MassAvg := StrToMass(str);
      end;
    end;
  finally
    sl.Free;
  end;
end;

function TAminoAcidList.GetMass: double;
var
  i: integer;
begin
  Result := 0.0;
  for i := 0 to Count - 1 do begin
    if not (Items[i] is TMod) then begin
      Result := Result + Items[i].Mass;
    end;
  end;
end;

{ TSequence }

constructor TSequence.Create;
begin
  inherited Create;
  fAAs := TAminoAcidList.Create;
  fMods := TModList.Create;
end;

procedure TSequence.SetMassType(Value: TMassType);
begin
  fAAs.MassType := Value;
  fMods.MassType := Value;
  fMassType := Value;
end;

function TSequence.GetSequence: string;
var
  i: integer;
begin
  Result := '';
  for i := 0 to Count - 1 do
    Result := Result + Items[i].NameChar;
end;

function TSequence.GetAAs: string;
var
  i: integer;
begin
  Result := '';
  for i := 0 to Count - 1 do
    if (Items[i] is TAminoAcid) then
      Result := Result + Items[i].NameChar;
end;

procedure TSequence.SetSequence(Value: string);
var
  i, cn: integer;
  c: Char;
  m: TMass;
begin
  Clear;
  cn := System.Length(Value);
  for i := 1 to cn do begin
    c := Value[i];
    m := fAAs.Find(c);
    if Assigned(m) then
      Add(m)
    else begin
      m := fMods.Find(c);
      if Assigned(m) then
        Add(m)
      else
        raise EInvalidSequence.CreateFmt('Invalid amino acid or modification "%s"', [c]);
    end;
  end;
end;

function TSequence.GetMass: double;
var
  i: integer;
  m: TMod;
  bct, bnt: boolean;
begin
  Result := inherited GetMass;

  for i := 0 to Count - 1 do begin
    if (Items[i] is TMod) then begin
      m := TMod(fMods.Find(Items[i].NameChar));
      if Assigned(m) and not m.Fixed then
        Result := Result + m.Mass;
    end else begin
      m := fMods.GetFromAA(Items[i].NameChar);
      if Assigned(m) and m.Fixed then
        Result := Result + m.Mass;
    end;
  end;

  if GetLength > 0 then begin
    bct := false; bnt := false;
    // Terminal modifications
    for i := 0 to fMods.Count - 1 do begin
      m := TMod(fMods[i]);
      if m.Selected then begin
        if not bct and (tC in m.Termini) then begin
          Result := Result + m.Mass;
          bct := true;
        end;
        if not bnt and (tN in m.Termini) then begin
          Result := Result + m.Mass;
          bnt := true;
        end;
      end;
    end;
  end;
end;

function TSequence.GetLength: integer;
var
  i: integer;
begin
  Result := 0;
  for i := 0 to Count - 1 do
    if (Items[i] is TAminoAcid) then
      Inc(Result);
end;

destructor TSequence.Destroy;
begin
  fMods.DeleteAll;
  fAAs.DeleteAll;
  fMods.Free;
  fAAs.Free;
  inherited;
end;

function TSequence.GetAAPos(const p: integer): integer;
var
  i: integer;
begin
  Result := 0;
  if (p = 0) or (Count = 0) then Exit;
  for i := 0 to Min(p-1, Count-1) do
    if Items[i] is TAminoAcid then
      Inc(Result);
end;

procedure TSequence.LoadFromFasta(const AFilename: string);
var
  sl: TStringList;
  str: string;
begin
  Clear;
  sl := TStringList.Create;
  try
    sl.LoadFromFile(AFilename);
    if sl.Count <> 2 then
      raise EUnknownFileFormat.Create('Unknown FASTA file format');
    str := sl[0];
    if (str = '') or (str[1] <> '>') then
      raise EUnknownFileFormat.Create('Unknown FASTA file format');
    fName := str;
    Sequence := sl[1];
  finally
    sl.Free;
  end;
end;

procedure TSequence.SaveToFasta(const AFilename: string);
var
  sl: TStringList;
begin
  sl := TStringList.Create;
  try
    sl.Add('>' + Name);
    sl.Add(GetAAs);
    sl.SaveToFile(AFilename);
  finally
    sl.Free;
  end;
end;

function TSequence.GetName: string;
begin
  if fName <> '' then
    Result := fName
  else
    Result := 'Unknown';
end;

end.
